/**
 * Copyright (c) 2020 Bubble, Inc.  All rights reserved.
 * For personal (non-commercial) use, see license: https://getbubblenow.com/bubble-license/
 */
package bubble.dao.device;

import bubble.dao.account.AccountDAO;
import bubble.dao.account.AccountOwnedEntityDAO;
import bubble.dao.app.AppDataDAO;
import bubble.model.device.BubbleDeviceType;
import bubble.model.device.Device;
import bubble.server.BubbleConfiguration;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.hibernate.criterion.Order;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;

import javax.transaction.Transactional;
import java.io.File;
import java.util.List;

import static bubble.ApiConstants.HOME_DIR;
import static bubble.model.device.Device.UNINITIALIZED_DEVICE_LIKE;
import static bubble.model.device.Device.newUninitializedDevice;
import static org.cobbzilla.util.daemon.ZillaRuntime.background;
import static org.cobbzilla.util.io.FileUtil.abs;
import static org.cobbzilla.util.io.FileUtil.touch;

@Repository @Slf4j
public class DeviceDAO extends AccountOwnedEntityDAO<Device> {

    public static final File VPN_REFRESH_USERS_FILE = new File(HOME_DIR, ".algo_refresh_users");

    private static short SPARE_DEVICES_PER_ACCOUNT_MAX = 10;
    private static short SPARE_DEVICES_PER_ACCOUNT_THRESHOLD = 5;

    @Autowired private BubbleConfiguration configuration;
    @Autowired private AppDataDAO dataDAO;

    @Override public Order getDefaultSortOrder() { return ORDER_CTIME_ASC; }

    public Device findByAccountAndName(String accountUuid, String name) {
        return findByUniqueFields("account", accountUuid, "name", name);
    }

    public Device findByNetworkAndName(String networkUuid, String name) {
        return findByUniqueFields("network", networkUuid, "name", name);
    }

    @Override public Object preCreate(Device device) {
        if (device.uninitialized()) {
            device.setDeviceType(BubbleDeviceType.uninitialized);
        }
        return super.preCreate(device);
    }

    @Transactional
    @Override public Device create(@NonNull final Device device) {
        if (device.uninitialized()) return super.create(device);

        final var accountUuid = device.getAccount();
        final var uninitializedDevices = findByAccountAndUninitialized(accountUuid);

        if (uninitializedDevices.isEmpty()) {
            log.info("create: no uninitialized devices for account " + accountUuid);
            background(() -> {
                if (!configuration.getBean(AccountDAO.class).findByUuid(accountUuid).isRoot()) {
                    ensureAllSpareDevices(accountUuid, device.getNetwork());
                }
            });
            // just create the device now:
            device.initTotpKey();
            device.initDeviceType();
            final var d = super.create(device);
            refreshVpnUsers();
            return d;
        } else {
            final var d = uninitializedDevices.get(0);
            d.initialize(device);
            return update(d);
            // update will do refreshVpnUsers
        }
    }

    @Override @NonNull public Device update(@NonNull final Device updateRequest) {
        final var old = findByUuid(updateRequest.getUuid());

        if (old != null && old.uninitialized()) updateRequest.initTotpKey();
        updateRequest.initDeviceType();

        final var updated = super.update(updateRequest);
        if (old.uninitialized() && !updated.uninitialized()
                && findByAccountAndUninitialized(old.getAccount()).size() < SPARE_DEVICES_PER_ACCOUNT_THRESHOLD) {
            background(() -> ensureAllSpareDevices(old.getAccount(), old.getNetwork()));
        }

        refreshVpnUsers();
        return updated;
    }

    @Override public void delete(String uuid) {
        final Device device = findByUuid(uuid);
        if (device != null) {
            dataDAO.deleteDevice(uuid);
            super.delete(uuid);
            if (device.uninitialized() && findByAccountAndUninitialized(device.getAccount()).size()
                                          < SPARE_DEVICES_PER_ACCOUNT_THRESHOLD) {
                background(() -> ensureAllSpareDevices(device.getAccount(), device.getNetwork()));
            }
            refreshVpnUsers();
        }
    }

    @Override public void forceDelete(String uuid) { super.delete(uuid); }

    @Transactional
    public synchronized boolean ensureAllSpareDevices(@NonNull final String account, @NonNull final String network) {
        final var currentSpareDevices = findByAccountAndUninitialized(account);
        boolean newDevicesCreated = false;
        for (var i = currentSpareDevices.size(); i < SPARE_DEVICES_PER_ACCOUNT_MAX; i++) {
            super.create(newUninitializedDevice(network, account));
            newDevicesCreated = true;
        }
        return newDevicesCreated;
    }

    public void refreshVpnUsers() {
        log.info("ensureSpareDevice: refreshing VPN users by touching: "+abs(VPN_REFRESH_USERS_FILE));
        touch(VPN_REFRESH_USERS_FILE);
    }

    public List<Device> findByAccountAndUninitialized(String accountUuid) {
        return findByFieldEqualAndFieldLike("account", accountUuid, "name", UNINITIALIZED_DEVICE_LIKE);
    }

}
